"""Deserializers for primitive types."""

from functools import partial
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    cast,
)

import attr

from . import pretty_printer
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .types import DeserializableType, Primitive, T, Tag


@attr.s(auto_attribs=True, frozen=True)
class PrimitiveDeserializer(Deserializer[T]):
    """Deserializer for a primitive type."""

    _deserializer: Callable[[Primitive], T]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """Deserialize a primitive value."""
        try:
            return Ok(self._deserializer(value))  # type: ignore[misc, call-arg]
        except _PrimitiveError as e:
            return Err(cs, str(e))


@attr.s(auto_attribs=True, frozen=True)
class PrimitiveDeserializerFactory(DeserializerFactory):
    """Create deserializer for a primitive type."""

    _bool_strings: Optional[Mapping[str, bool]] = None

    def mk_ds_internal(
        self, _: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified primitive type."""
        if self._bool_strings is not None:
            deserialize_int = _deserialize_int_from_string
            deserialize_float = _deserialize_float_from_string
            deserialize_bool: Callable[[Primitive], bool] = partial(
                _deserialize_bool_from_string,
                {s for s, v in self._bool_strings.items() if v},
                {s for s, v in self._bool_strings.items() if not v},
            )
        else:
            deserialize_int = _deserialize_int
            deserialize_float = _deserialize_float
            deserialize_bool = _deserialize_bool
        ds = cast(
            Dict[DeserializableType, Callable[[Primitive], T]],
            {
                Any: lambda x: x,
                str: _deserialize_str,
                int: deserialize_int,
                float: deserialize_float,
                dict: _deserialize_dict,
                list: _deserialize_list,
                bool: deserialize_bool,
                type(None): _deserialize_none,
            },
        )
        try:
            return Ok(Ok(PrimitiveDeserializer(ds[type_.t])))
        except (KeyError, TypeError):
            # A TypeError is raised if the type is unhashable.
            return Err(cs, "type mismatch")


@attr.s(auto_attribs=True, frozen=True)
class _PrimitiveError(Exception):
    msg: str


def _deserialize_str(value: Primitive) -> str:
    """Deserialise a primitive as a string."""
    return _to_primitive(value, str)


def _deserialize_int(value: Primitive) -> int:
    """
    Deserialise a primitive as an int.

    Union[bool, int] will be collapsed into int, so special-case bools.
    TODO - deserialize with best match
    """
    if isinstance(value, bool):
        return value
    return _to_primitive(value, int)


def _deserialize_int_from_string(value: Primitive) -> int:
    value = _assert_string(value, int).lower()
    try:
        return int(value)
    except ValueError as e:
        raise _PrimitiveError(f"could not coerce value '{value}' to type 'int'") from e


def _deserialize_float(value: Primitive) -> float:
    """
    Deserialize a primitive as a float.

    Both ints and floats are acceptable input values.
    """
    try:
        return _to_primitive(value, int)
    except _PrimitiveError:
        return _to_primitive(value, float)


def _deserialize_float_from_string(value: Primitive) -> float:
    value = _assert_string(value, float).lower()
    try:
        return float(value)
    except ValueError as e:
        raise _PrimitiveError(
            f"could not coerce value '{value}' to type 'float'"
        ) from e


def _deserialize_dict(value: Primitive) -> Dict[str, Any]:
    """Deserialize a primitive as an untyped dict."""
    return _to_primitive(value, dict)


def _deserialize_list(value: Primitive) -> List[Any]:
    """Deserialize a primitive as an untyped list."""
    return _to_primitive(value, list)


def _deserialize_bool(value: Primitive) -> bool:
    """Deserialize a primitive as a bool."""
    return _to_primitive(value, bool)


def _deserialize_bool_from_string(
    true_values: AbstractSet[str], false_values: AbstractSet[str], value: Primitive
) -> bool:
    """Deserialize a string as a bool."""
    value = _assert_string(value, bool).lower()
    if value in true_values:
        return True
    elif value in false_values:
        return False
    else:
        raise _PrimitiveError(
            "could not coerce value '{}' to type 'bool' (expected one of {})".format(
                value, set(true_values | false_values)
            )
        )


def _deserialize_none(value: Primitive) -> None:
    """Deserialize a primitive as None."""
    if value is None:
        return value
    _raise_type_mismatch(value, type(None))


S = TypeVar("S")


def _to_primitive(value: Primitive, primitive_t: Type[S]) -> S:
    if isinstance(value, primitive_t):
        return primitive_t.__call__(value)
    _raise_type_mismatch(value, primitive_t)


def _assert_string(value: Primitive, type_: type) -> str:
    if isinstance(value, str):
        return value
    if isinstance(value, (bool, int, float)):
        raise _PrimitiveError(
            "type mismatch - expected 'str' (as coerce_strings is set)"
        )
    _raise_type_mismatch(value, type_)


def _raise_type_mismatch(value: Primitive, type_: type) -> NoReturn:
    raise _PrimitiveError(
        f"type mismatch - expected '{pretty_printer.print_type_name(type_)}'"
    )



"""Common utility functions for dealing with generic types."""

import contextlib
from collections import abc
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
)

from typing_extensions import Literal, TypedDict

from .types import DeserializableType, LiteralMeta, TypedDictMetas


def get_base_of_generic_type(type_: DeserializableType) -> Optional[Type[Any]]:
    """Return the generic type of which a type_ is an instance."""
    type_ = _normalize_type(type_)
    for attribute in ("__origin__", "__class__"):
        if hasattr(type_, attribute):
            return _coalesce_type(getattr(type_, attribute))
    return None  # pragma: no cover


def get_type_parameters(from_type: DeserializableType) -> List[Type[Any]]:
    """Extract the type parameters from a parametrized type."""

    def extract_type_parameters(type_: DeserializableType) -> List[Type[Any]]:
        type_ = _normalize_type(type_)
        with contextlib.suppress(AttributeError):
            if type_.__args__ is not None:  # type: ignore[union-attr]
                return list(type_.__args__)  # type: ignore[union-attr]
        with contextlib.suppress(AttributeError):
            if type_.__values__ is not None:  # type: ignore[union-attr]
                return list(type_.__values__)  # type: ignore[union-attr]
        return []

    from_type = _normalize_type(from_type)
    type_params = extract_type_parameters(_normalize_type(from_type))
    if getattr(from_type, "__class__", None) is LiteralMeta:
        # Special-case Literal types, as Literal[()] != Literal[]
        return type_params
    return type_params if type_params != [()] else []


def _normalize_type(type_: Any) -> Any:
    type_ = _coalesce_type(type_)
    try:
        return {
            AbstractSet: AbstractSet[Any],
            Callable: Callable[..., Any],
            Dict: Dict[str, Any],
            FrozenSet: FrozenSet[Any],
            Iterable: Iterable[Any],
            Iterator: Iterator[Any],
            List: List[Any],
            Mapping: Mapping[str, Any],
            MutableMapping: Mapping[str, Any],
            MutableSequence: MutableSequence[Any],
            Sequence: Sequence[Any],
            Set: Set[Any],
            Tuple: Tuple[Any, ...],
        }[type_]
    except (KeyError, TypeError):
        return type_


def _coalesce_type(type_: Any) -> Any:
    try:
        return {
            abc.Callable: Callable,
            abc.Iterable: Iterable,
            abc.Iterator: Iterator,
            abc.Mapping: Mapping,
            abc.MutableMapping: MutableMapping,
            abc.MutableSequence: MutableSequence,
            abc.Sequence: Sequence,
            abc.Set: Set,
            dict: Dict,
            frozenset: FrozenSet,
            list: List,
            LiteralMeta: Literal,
            set: Set,
            tuple: Tuple,
            **{meta: TypedDict for meta in TypedDictMetas},
        }[type_]
    except (KeyError, TypeError):
        return type_



"""Deserializer for a newtype alias."""

import attr

from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .types import T, Tag


@attr.s(auto_attribs=True, frozen=True)
class NewTypeDeserializerFactory(DeserializerFactory):
    """Create a newtype deserializer."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified newtype."""
        if not getattr(type_.t, "__qualname__", None) == "NewType.<locals>.new_type":
            return Err(cs, "type mismatch")
        return Ok(factory.mk_ds(cs, Tag(type_.t.__supertype__))())  # type: ignore



"""
Experimental features which are not guaranteed to be supported.

Breaking changes to these features do not trigger a major version bump.
"""

from .metadata import DICT, LIST, VALUE, From, Keyed, With_, from_, keyed, with_

__all__ = ["DICT", "LIST", "VALUE", "From", "Keyed", "With_", "from_", "keyed", "with_"]



"""Utilities for dealing with terramare's error types."""

import abc
import itertools
import json
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import attr
import result
from result import Ok

from .iterator_utils import round_robin, round_robin_group
from .pretty_printer import print_type_name
from .tree import Tree

__all__ = [
    "ContextStack",
    "Err",
    "Ok",  # re-export
    "Result",
    "add_err",
    "collect_all",
    "collect_any",
    "collect_any_it",
    "join",
    "unwrap",
]

T = TypeVar("T")


class _Error(abc.ABC):
    @abc.abstractmethod
    def __add__(self, other: object) -> "_Error":
        """Combine two errors."""

    @abc.abstractmethod
    def map_(self, fn: Callable[["_LeafError"], None]) -> None:
        """Map a function over all leaf errors."""


@attr.s(auto_attribs=True, frozen=True)
class _LeafError(_Error):
    path: Tuple[str, ...]
    msg: str

    def __add__(self, other: object) -> "_Error":
        raise NotImplementedError  # pragma: no cover

    def map_(self, fn: Callable[["_LeafError"], None]) -> None:
        fn(self)


@attr.s(auto_attribs=True, frozen=True)
class _CompositeError(_Error):
    errors: List[Union[_LeafError, "_CompositeError"]]

    def __add__(self, other: object) -> "_Error":
        if isinstance(other, _LeafError):
            return _CompositeError(self.errors + [other])
        elif isinstance(other, _CompositeError):  # pragma: no cover
            return _CompositeError(self.errors + other.errors)
        raise NotImplementedError  # pragma: no cover

    def map_(self, fn: Callable[[_LeafError], None]) -> None:
        for error in self.errors:
            error.map_(fn)


@attr.s(auto_attribs=True)
class ContextStack:
    """Stack data structure for providing error context."""

    _stack: Tuple[str, ...]

    @staticmethod
    def new() -> "ContextStack":
        """Create a ContextStack."""
        return ContextStack(())

    def index(self, index: Union[int, str]) -> "ContextStack":
        """Push an index onto the context stack."""
        return self._push(json.dumps(index))

    def type_index(self, type_: Any) -> "ContextStack":
        """Push a type index onto the context stack."""
        return self._push(print_type_name(type_))

    def parameter_index(self, index: Union[int, str], type_: Any) -> "ContextStack":
        """Push a type parameter index onto the context stack."""
        return self._push(f"{json.dumps(index)}: {print_type_name(type_)}")

    def alternative(self, alternative: Union[int, str]) -> "ContextStack":
        """Push an alternative onto the context stack."""
        return self._push(f"• {alternative}")

    def _push(self, context: str) -> "ContextStack":
        return ContextStack(self._stack + (context,))

    def make_err(self, msg: str) -> _LeafError:
        """Create an error using the current context."""
        return _LeafError(self._stack, msg)


Result = result.Result[T, Iterator[_Error]]


def Err(cs: ContextStack, msg: str) -> Result[T]:
    """Create an error with context."""
    return result.Err(iter([cs.make_err(msg)]))


def add_err(err: _Error, errs: Iterator[_Error]) -> Iterator[_Error]:
    """Add an error to the first set of errors produced by an iterator."""
    try:
        yield next(errs) + err
    except StopIteration:  # pragma: no cover
        yield err
    yield from errs


Fn = Callable[[], Result[Any]]
Collectable = Union[Iterator[Fn], Iterable[Fn]]


def collect_all(fns: Collectable) -> Result[List[Any]]:
    """
    Lazily evaluate a series of functions and return the results.

    If any function returns an error, return an error.
    """
    vs = []
    fns = iter(fns)
    for fn in fns:
        v = fn()
        if v.is_err():
            # Return early if any ds fails.
            # Create an iterator over the rest of the errors.
            extra_errs = (fn().map(lambda _: iter(())).value for fn in fns)
            # Yield errors breadth-first.
            return result.Err(
                round_robin(itertools.chain((v.unwrap_err(),), extra_errs))
            )
        vs.append(v.unwrap())
    return Ok(vs)


def collect_any_it(fns: Collectable) -> Result[Iterator[Any]]:
    """
    Lazily evaluate a series of functions and return the results.

    If any function returns ok, return ok.
    """
    errs = []
    fns = iter(fns)
    for fn in fns:
        v = fn()
        if v.is_ok():
            # Return early if any ds succeeds.
            # Create an iterator over the rest of the successes.
            extra_vs = map(
                result.Result.unwrap, filter(result.Result.is_ok, (fn() for fn in fns))
            )
            # Yield errors breadth-first.
            return Ok(itertools.chain((v.unwrap(),), extra_vs))
        errs.append(v.unwrap_err())
    return result.Err(map(_CompositeError.__call__, round_robin_group(errs)))


def collect_any(fns: Collectable) -> Result[Any]:
    """
    Lazily evaluate a series of functions and return the results.

    If any function returns ok, return ok.
    """
    result = collect_any_it(fns)
    if result.is_ok():
        return Ok(next(result.unwrap()))
    return result


def unwrap(exc: Callable[[Tree[str, str]], Exception], value: Result[T]) -> T:
    """Unwrap a result, raising an exception if it is not Ok."""
    if value.is_ok():
        return value.unwrap()
    raise exc(_treeify(value.unwrap_err()))


def join(value: Result[Result[T]]) -> Result[T]:
    """Flatten a Result[Result] into a Result."""
    if value.is_ok():
        return value.unwrap()
    return result.Err(value.unwrap_err())


def _treeify(
    errs: Iterator[_Error], max_errors: Optional[int] = None
) -> Tree[str, str]:
    tree: Tree[str, str] = Tree()

    def add(err: _LeafError) -> None:
        tree[err.path] = err.msg

    for err, _ in zip(
        errs, range(max_errors) if max_errors is not None else itertools.repeat(0)
    ):
        err.map_(add)
    return tree



"""Common types and re-exports."""

from typing import Any, Generic, Mapping, Sequence, TypeVar, Union

import attr
from typing_extensions import Literal

TypedDictMetas = set()

# pylint: disable=unused-import
try:  # pragma: no cover
    from typing import _TypedDictMeta  # type: ignore[attr-defined]

    TypedDictMetas.add(_TypedDictMeta)

except ImportError:  # pragma: no cover
    pass

# pylint: disable=unused-import
try:  # pragma: no cover
    from typing_extensions import _TypedDictMeta  # type: ignore[attr-defined]

    TypedDictMetas.add(_TypedDictMeta)
except ImportError:  # pragma: no cover
    pass

try:
    from typing_extensions import _Literal  # type: ignore[attr-defined]

    LiteralMeta = _Literal
except ImportError:  # pragma: no cover
    LiteralMeta = Literal

T = TypeVar("T")


NotNonePrimitive = Union[str, int, float, Mapping, Sequence, bool]
Primitive = Union[NotNonePrimitive, None]

DeserializableType = Any


class TerramareError(Exception):
    """Base class for exceptions raised by terramare."""


@attr.s(auto_attribs=True, frozen=True)
class Tag(Generic[T]):
    """Tag type to satisfy the type checker."""

    t: DeserializableType



"""Terramare-specific metadata for attrs classes."""

import enum
from typing import AbstractSet, Any, Callable, Generic, Mapping, cast

import attr
from typing_extensions import Final

from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .keyed_deserializer import K, KeyedDeserializerFactory, S
from .pretty_printer import print_type_name
from .types import DeserializableType, T, Tag

_METADATA_KEY: Final = "__terramare_terramare__"


class Metadata:
    """Base class for terramare metadata."""


@attr.s(auto_attribs=True, frozen=True)
class MetadataCollection:
    """Mapping of DeserializableTypes to terramare metadata."""

    _metadata: Mapping[DeserializableType, Metadata]

    def __getitem__(self, k: DeserializableType) -> Metadata:
        """Retrieve the metadata for a type."""
        try:
            if k in self._metadata:
                return self._metadata[k]
        except TypeError:
            # A TypeError is raised if k is not hashable.
            pass
        if hasattr(k, _METADATA_KEY):
            pair = getattr(k, _METADATA_KEY)
            if not isinstance(pair, tuple) or len(pair) != 2:
                raise TypeError("Metadata type mismatch")
            t, t_metadata = pair
            if not isinstance(t_metadata, Metadata):
                raise TypeError("Metadata type mismatch")
            if t == k:
                return t_metadata
        raise KeyError(k)


@attr.s(auto_attribs=True, frozen=True)
class MetadataDeserializerFactory(DeserializerFactory):
    """Create a deserializer from metadata."""

    _metadata: MetadataCollection
    _keyed_ds_factory: KeyedDeserializerFactory[Any, Any, Any]

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified type."""
        try:
            metadata = self._metadata[type_.t]
        except KeyError:
            return Err(cs, "no metadata supplied")
        if isinstance(metadata, From):
            return Err(cs, "no metadata supplied")
        if isinstance(metadata, With_):
            return Ok(
                cast(
                    Result[Deserializer[T]], factory.mk_ds(cs, Tag(metadata.with_()))()
                )
            )
        if isinstance(metadata, Keyed):
            return Ok(
                self._keyed_ds_factory.mk_ds_internal(
                    factory, cs, metadata.key, metadata.mapping()
                )
            )
        raise NotImplementedError(
            f"unexpected metadata type: {print_type_name(metadata)}"
        )  # pragma: no cover


@attr.s(auto_attribs=True, frozen=True)
class Keyed(Generic[K, S], Metadata):
    """
    Metadata to deserialize into a type determined by the value of a key in the input dictionary.

    See the `keyed` decorator for more details.
    """

    key: K
    mapping: Callable[[], Mapping[S, DeserializableType]]


def keyed(
    key: K, mapping: Callable[[], Mapping[S, DeserializableType]]
) -> Callable[[type], type]:
    """
    Metadata to deserialize into a type determined by the value of a key in the input dictionary.

    :param `key`: Use the value of this field to determine the target type.
    :param `mapping`: Callable returning the mapping of possible key values to target types.

    Example usage:

    >>> from typing import Any
    >>> import attr
    >>> import terramare
    >>>
    >>> @keyed("type", lambda: {0: IntVariant, 1: StrVariant})
    ... class Variant:
    ...     pass
    >>>
    >>> @attr.s(auto_attribs=True)
    ... class IntVariant(Variant):
    ...     integer: int
    >>>
    >>> @attr.s(auto_attribs=True)
    ... class StrVariant(Variant):
    ...     string: str
    >>>
    >>> terramare.deserialize_into(Variant, {"type": 0, "integer": 1})
    IntVariant(integer=1)
    >>>
    >>> terramare.deserialize_into(Variant, {"type": 1, "string": "string"})
    StrVariant(string='string')

    """
    return _set_metadata_fn(Keyed(key, mapping))


Decorator = Callable[[Any], Any]


@enum.unique
class _FromType(enum.Enum):
    DICT = "dict"
    LIST = "list"
    VALUE = "value"


FromType = AbstractSet[_FromType]

DICT: Final = frozenset({_FromType.DICT})
LIST: Final = frozenset({_FromType.LIST})
VALUE: Final = frozenset({_FromType.VALUE})


@attr.s(auto_attribs=True, frozen=True)
class From(Metadata):
    """
    Specify the primitive types from which a class can be deserialized.

    See the `from_` decorator for more details.
    """

    from_: FromType


def from_(from_: FromType) -> Decorator:
    """
    Specify the primitive types from which a class can be deserialized.

    :param `from_`: Set of deserialization methods.

    Example usage:

    >>> import terramare
    >>>
    >>> class Int:
    ...    def __init__(self, x: int):
    ...        self.x = x
    ...
    ...    def __repr__(self) -> str:
    ...        return f"Int({self.x})"
    >>>
    >>> # Deserializable from a dict only.
    >>> @from_(DICT)
    ... class FromDict(Int):
    ...     pass
    >>>
    >>> terramare.deserialize_into(FromDict, {"x": 1})
    Int(1)
    >>>
    >>> # Deserializable from a list only.
    >>> @from_(LIST)
    ... class FromList(Int):
    ...     pass
    >>>
    >>> terramare.deserialize_into(FromList, [2])
    Int(2)
    >>>
    >>> # Deserializable from a single value only.
    >>> @from_(VALUE)
    ... class FromValue(Int):
    ...     pass
    >>>
    >>> terramare.deserialize_into(FromValue, 3)
    Int(3)
    >>>
    >>> # Deserializable from a dict or a list.
    >>> @from_(DICT | LIST)
    ... class FromDictOrList(Int):
    ...     pass
    >>>
    >>> terramare.deserialize_into(FromDictOrList, {"x": 4})
    Int(4)
    >>> terramare.deserialize_into(FromDictOrList, [5])
    Int(5)
    """
    return _set_metadata_fn(From(from_))


@attr.s(auto_attribs=True, frozen=True)
class With_(Metadata):
    """
    Specify a function to use to construct instances of a type.

    See the `with_` decorator for more details.
    """

    with_: Callable[[], DeserializableType]


def with_(with_: Callable[[], DeserializableType]) -> Decorator:
    """
    Specify a function to use to construct instances of a type.

    :param with_: Callable returning the construction function to use.

    Example usage:

    >>> import terramare
    >>>
    >>> @with_(lambda: Example.new)
    ... class Example:
    ...     def __init__(self, x: int):
    ...         self.x = x
    ...
    ...     def __repr__(self) -> str:
    ...         return f"Example({self.x})"
    ...
    ...     @staticmethod
    ...     @from_(VALUE)
    ...     def new(y: int) -> "Example":
    ...         return Example(y + 1)
    >>>
    >>> terramare.deserialize_into(Example, 1)
    Example(2)
    """
    return _set_metadata_fn(With_(with_))


def _set_metadata_fn(metadata: Metadata) -> Decorator:
    def inner(t: type) -> type:
        setattr(t, _METADATA_KEY, (t, metadata))
        return t

    return inner



"""Default metadata for selected standard library types."""


import datetime
import ipaddress
import pathlib
import re
from functools import partial
from typing import Any, Callable, Mapping, Optional, Pattern, TypeVar

import attr
import dateutil.parser
from typing_extensions import Final

from terramare.metadata import Metadata
from terramare.types import DeserializableType

from . import metadata


@attr.s(auto_attribs=True, frozen=True)
class StdlibDeserializationError(Exception):
    """Exception type raised on failure by standard library deserializers."""

    msg: str


@metadata.from_(metadata.VALUE)
def _make_datetime(s: str) -> datetime.datetime:
    try:
        return dateutil.parser.isoparse(s)
    except ValueError as e:
        raise StdlibDeserializationError("invalid ISO-8601 datetime") from e


@metadata.from_(metadata.VALUE)
def _make_regex(s: str) -> Pattern[Any]:
    try:
        return re.compile(s)
    except re.error as e:
        raise StdlibDeserializationError(f"invalid regex: {e}") from e


@metadata.from_(metadata.VALUE)
def _make_ipv4_address(s: str) -> ipaddress.IPv4Address:
    try:
        return ipaddress.IPv4Address(s)
    except ValueError as e:
        raise StdlibDeserializationError(f"invalid IPv4 address: {e}") from e


@metadata.from_(metadata.VALUE)
def _make_ipv6_address(s: str) -> ipaddress.IPv6Address:
    try:
        return ipaddress.IPv6Address(s)
    except ValueError as e:
        raise StdlibDeserializationError(f"invalid IPv6 address: {e}") from e


T = TypeVar("T")


def _make_str_fn(path_t: Callable[[str], T]) -> Callable[[str], T]:
    @metadata.from_(metadata.VALUE)
    def _make_path(s: str) -> T:
        return path_t(s)

    return _make_path


DEFAULT_METADATA: Final[Mapping[DeserializableType, Optional[Metadata]]] = {
    datetime.datetime: metadata.With_(lambda: _make_datetime),
    Pattern: metadata.With_(lambda: _make_regex),
    ipaddress.IPv4Address: metadata.With_(lambda: _make_ipv4_address),
    ipaddress.IPv6Address: metadata.With_(lambda: _make_ipv6_address),
    **{
        t: metadata.With_(partial(_make_str_fn, t))
        for t in {
            pathlib.Path,
            pathlib.PosixPath,
            pathlib.WindowsPath,
            pathlib.PurePath,
            pathlib.PurePosixPath,
            pathlib.PureWindowsPath,
        }
    },
}



"""Pretty-printer for error messages."""

import contextlib
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from typing_extensions import Literal

from . import type_utils
from .types import DeserializableType, Primitive


def print_primitive(
    value: Primitive,
    short: bool = False,
    max_string_length: int = 16,
    max_dict_length: int = 3,
    max_list_length: int = 3,
) -> str:
    """Pretty-print a primitive value."""
    if isinstance(value, str):
        if len(value) > max_string_length:
            value = value[: max_string_length - 3] + "..."
        return '"{}"'.format(value)
    if not value:
        return str(value)
    if isinstance(value, dict):
        if short:
            return "{...}"
        else:
            elts = [
                print_primitive(k) + ": " + print_primitive(v, short=True)
                for k, v in value.items()
            ]
            if len(elts) > max_dict_length:
                elts = elts[: max_dict_length - 1] + ["..."]
            return "{{{}}}".format(", ".join(elts))
    if isinstance(value, list):
        if short:
            return "[...]" if value else "[]"
        else:
            elts = [print_primitive(v, short=True) for v in value]
            if len(elts) > max_list_length:
                elts = elts[: max_list_length - 1] + ["..."]
            return "[{}]".format(", ".join(elts))
    return str(value)


def print_type_name(type_: Any) -> str:
    """Return the best available user-facing name for a type."""
    overrides: Dict[DeserializableType, str] = {
        dict: "dict",
        frozenset: "frozenset",
        set: "set",
        list: "list",
        tuple: "tuple",
    }
    if type_ in overrides:
        return overrides[type_]
    base = type_utils.get_base_of_generic_type(type_)
    if base in {
        Callable,
        Dict,
        FrozenSet,
        Iterable,
        Iterator,
        Literal,
        List,
        Mapping,
        MutableMapping,
        MutableSequence,
        Optional,
        Sequence,
        Set,
        Tuple,
        Union,
    }:
        return (
            _get_type_name(base)
            + "["
            + ", ".join(
                [print_type_name(t) for t in type_utils.get_type_parameters(type_)]
            )
            + "]"
        )
    return _get_type_name(type_)


def _get_type_name(type_: Any) -> str:
    overrides = {Any: "Any", Ellipsis: "...", Literal: "Literal", Union: "Union"}
    if type_ in overrides:
        return overrides[type_]
    with contextlib.suppress(AttributeError):  # pragma: no cover TODO
        # pylint: disable=protected-access
        if type_._name is not None:
            return type_._name
    with contextlib.suppress(AttributeError):
        return type_.__name__
    with contextlib.suppress(AttributeError):
        return type_.__qualname__
    if hasattr(type_, "__origin__"):  # pragma: no cover TODO
        return _get_type_name(type_.__origin__)
    return str(type_)



"""Common utility functions for dealing with iterators."""

import itertools
from typing import (
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

T = TypeVar("T")
S = TypeVar("S")


def zip_strict(
    required_ts: Sequence[T],
    optional_ts: Sequence[T],
    var_ts: Optional[T],
    ss: Sequence[S],
) -> Iterator[Tuple[T, S]]:
    """
    Zip required, optional and variadic elements with a supplied sequence.

    Raises if:
      - the sequence contains fewer than the required number of elements;
      - the sequence contains more than the required + optional number of
        elements and no variadic element is supplied.
    """
    if len(ss) < len(required_ts):
        raise ValueError(
            "too few elements ({}) - expected at least {}".format(
                len(ss), len(required_ts)
            )
        )
    if var_ts is not None:
        return zip(
            itertools.chain(
                iter(required_ts), iter(optional_ts), itertools.repeat(var_ts)
            ),
            ss,
        )
    ts = [*required_ts, *optional_ts]
    if len(ss) > len(ts):
        raise ValueError(
            "too many elements ({}) - expected at most {}".format(len(ss), len(ts))
        )
    return zip(ts, ss)


K = TypeVar("K")


def zip_strict_dict(
    required_ts: Mapping[K, T],
    optional_ts: Mapping[K, T],
    var_ts: Optional[T],
    ss: Mapping[K, S],
) -> Dict[K, Tuple[T, S]]:
    """
    Zip required, optional, and variadic key/value pairs with a supplied mapping.

    Raises if:
      - the mapping is missing required keys;
      - the mapping contains keys not present in the set of required or optional
        keys and no variadic element is supplied.
    """

    def show_if(msg: str, keys: Set[K]) -> Optional[str]:
        return msg.format(", ".join(map(str, keys))) if keys else None

    missing_keys = show_if("missing keys '{}'", set(required_ts) - set(ss))
    ts = {**required_ts, **optional_ts}
    unexpected_keys = None
    if var_ts is None:
        unexpected_keys = show_if("unexpected keys '{}'", set(ss) - set(ts))
    if missing_keys or unexpected_keys:
        raise ValueError(
            "key mismatch - {}".format(
                ", ".join((msg for msg in (missing_keys, unexpected_keys) if msg))
            )
        )
    # Either var_ts is set, or it is None and all keys in ss are present in ts.
    return {k: (ts.get(k, cast(T, var_ts)), ss[k]) for k in ss}


def divide_list(fn: Callable[[T], bool], ts: Sequence[T]) -> Tuple[List[T], List[T]]:
    """Divide a list into elements that meet a predicate and those that don't."""
    predicate_met: List[T] = []
    predicate_not_met: List[T] = []
    for t in ts:
        if fn(t):
            predicate_met.append(t)
        else:
            predicate_not_met.append(t)
    return (predicate_met, predicate_not_met)


def round_robin(iterators: Iterator[Iterator[T]]) -> Iterator[T]:
    """
    Round-robin yield from a list of iterators.

    Yield a value from each iterator in turn until all iterators are exhausted.
    For example:

    >>> list(round_robin(iter([iter([1,2,3]), iter(['a', 'b']), iter(['x'])])))
    [1, 'a', 'x', 2, 'b', 3]
    """
    its: Union[Iterator[Iterator[T]], List[Iterator[T]]] = iterators
    while its:
        new_its = []
        for it in its:
            try:
                yield next(it)
            except StopIteration:
                continue
            new_its.append(it)
        its = new_its


def round_robin_group(iterators: Sequence[Iterator[T]]) -> Iterator[List[T]]:
    """
    Yield lists containing one value from each of a list of iterators.

    For example:

    >>> list(round_robin_group([iter([1, 2, 3]), iter(['a', 'b']), iter(['x'])]))
    [[1, 'a', 'x'], [2, 'b'], [3]]
    """
    while iterators:
        group = []
        new_its = []
        for it in iterators:
            try:
                group.append(next(it))
            except StopIteration:
                continue
            new_its.append(it)
        if group:
            yield group
        iterators = new_its



"""Deserializer factory for a generic class."""

import inspect
import sys
import typing
from functools import partial
from inspect import Parameter
from typing import AbstractSet, Any, Callable, Dict, List, Sequence, Type, TypeVar, cast

import attr

from . import error_utils, iterator_utils, mappings, sequences, unions
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .metadata import DICT, LIST, VALUE, FromType, _FromType
from .types import DeserializableType, Primitive, T, Tag

S = TypeVar("S")


@attr.s(auto_attribs=True, frozen=True)
class ClassDeserializerFactory(DeserializerFactory):
    """Create deserializer for a class."""

    _enable_if: Callable[[DeserializableType], FromType] = lambda _: DICT | LIST | VALUE
    _handle_exception_ts: AbstractSet[Type[Exception]] = attr.ib(factory=frozenset)

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified class type."""

        def _mk_ds(name: str, fn: Any) -> Callable[[], Result[Deserializer[T]]]:
            return lambda: fn(
                self._handle_exception_ts, factory, cs.alternative(name), type_, params
            ).map(lambda ds: (name, ds))

        ds_fns = [
            _mk_ds(t.value, fn)
            for t, fn in {
                _FromType.DICT: _mk_dict_ds,
                _FromType.LIST: _mk_list_ds,
                _FromType.VALUE: _mk_value_ds,
            }.items()
            if t in self._enable_if(type_.t)
        ]

        if not hasattr(type_.t, "__call__"):
            return Err(cs, "type mismatch")
        if not ds_fns:
            return Err(cs, "not enabled for type")
        try:
            params = _get_parameters(type_.t)
        except _ParameterError as e:
            return Ok(Err(cs, e.msg))

        return Ok(
            error_utils.collect_any_it(ds_fns).map(
                lambda dss: unions.OneOfDeserializer(dict(dss))
            )
        )


def _mk_dict_ds(
    handle_exception_ts: AbstractSet[Type[Exception]],
    factory: DeserializerFactory,
    cs: ContextStack,
    type_: Tag[T],
    params: Sequence["_Parameter"],
) -> Result[Deserializer[T]]:
    non_kw_required = [p.name for p in params if p.is_required() and not p.is_kw()]
    # Non-keyword parameters can't be created easily pre-Python 3.8.
    if non_kw_required:  # pragma: no cover
        return Err(
            cs, "required non-keyword parameter(s) " + ", ".join(non_kw_required)
        )

    params = [p for p in params if p.is_kw()]
    result = _get_dss(factory, cs, type_, [dict, str], params)
    if result.is_err():
        return cast(Result[Deserializer[T]], result)
    dss = result.unwrap()

    var, non_var = iterator_utils.divide_list(_Parameter.is_var, params)
    param_dss = dict(zip((p.name for p in params), dss[2:]))
    return Ok(
        mappings.MappingDeserializer(
            dss[0],
            dss[1],
            {p.name: param_dss[p.name] for p in non_var if p.is_required()},
            {p.name: param_dss[p.name] for p in non_var if not p.is_required()},
            next(iter([param_dss[p.name] for p in var]), None),
            _mk_construct_fn(handle_exception_ts, lambda v: type_.t(**v)),
        )
    )


def _mk_list_ds(
    handle_exception_ts: AbstractSet[Type[Exception]],
    factory: DeserializerFactory,
    cs: ContextStack,
    type_: Tag[T],
    params: Sequence["_Parameter"],
) -> Result[Deserializer[T]]:
    non_pos_required = [p.name for p in params if p.is_required() and not p.is_pos()]
    if non_pos_required:
        return Err(
            cs, "required non-positional parameter(s) " + ", ".join(non_pos_required)
        )

    params = [p for p in params if p.is_pos()]
    result = _get_dss(factory, cs, type_, [list], params)
    if result.is_err():
        return cast(Result[Deserializer[T]], result)
    dss = result.unwrap()

    var, non_var = iterator_utils.divide_list(_Parameter.is_var, params)
    param_dss = dict(zip((p.name for p in params), dss[1:]))
    return Ok(
        sequences.SequenceDeserializer(
            dss[0],
            [param_dss[p.name] for p in non_var if p.is_required()],
            [param_dss[p.name] for p in non_var if not p.is_required()],
            next(iter([param_dss[p.name] for p in var]), None),
            _mk_construct_fn(handle_exception_ts, lambda v: type_.t(*v)),
        )
    )


@attr.s(auto_attribs=True, frozen=True)
class _ClassValueDeserializer(Deserializer[T]):
    """Deserialize a class from a single value."""

    _handle_exception_ts: AbstractSet[Type[Exception]]
    _construct_fn: Callable[[ContextStack, Any], Result[T]]
    _value_ds: Deserializer[Any]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        result = self._value_ds.ds(cs, value)
        if result.is_err():
            return cast(Result[T], result)
        return self._construct_fn(cs, result.unwrap())


def _mk_value_ds(
    handle_exception_ts: AbstractSet[Type[Exception]],
    factory: DeserializerFactory,
    cs: ContextStack,
    type_: Tag[T],
    params: Sequence["_Parameter"],
) -> Result[Deserializer[T]]:
    required, optional = iterator_utils.divide_list(_Parameter.is_required, params)
    if len(required) > 1 or len(required) + len(optional) < 1:
        return Err(
            cs,
            "unsupported number of parameters "
            + f"(required: {len(required)}, optional: {len(optional)})",
        )

    value_param = params[0]
    if not value_param.is_pos():
        return Err(cs, "non-positional first parameter")
    return _get_dss(factory, cs, type_, [], [value_param]).map(
        lambda dss: _ClassValueDeserializer(
            handle_exception_ts, _mk_construct_fn(handle_exception_ts, type_.t), dss[0]
        )
    )


def _get_dss(
    factory: DeserializerFactory,
    cs: ContextStack,
    type_: Tag[T],
    extra: Sequence[DeserializableType],
    params: Sequence["_Parameter"],
) -> Result[List[Deserializer[Any]]]:
    try:
        return error_utils.collect_all(
            [factory.mk_ds(cs, Tag(t)) for t in extra]
            + [
                factory.mk_ds(cs.parameter_index(p, t), Tag(t))
                for p, t in _get_type_hints(
                    type_.t, [p.parameter for p in params]
                ).items()
            ]
        )
    except _ParameterError as e:  # pragma: no cover
        # This branch is pretty hard to hit.
        return Err(cs, e.msg)


def _mk_construct_fn(
    handle_exception_ts: AbstractSet[Type[Exception]], fn: Callable[[Any], T]
) -> Callable[[ContextStack, Any], Result[T]]:
    def inner(cs: ContextStack, args: Any) -> Result[T]:
        try:
            return Ok(fn(args))
        # pylint: disable=catching-non-exception
        except tuple(handle_exception_ts) as e:
            return Err(cs, str(e))

    return inner


@attr.s(auto_attribs=True, frozen=True)
class _Parameter:
    parameter: Parameter

    @property
    def name(self) -> str:
        return self.parameter.name

    def is_var(self) -> bool:
        return self.parameter.kind in {
            Parameter.VAR_POSITIONAL,
            Parameter.VAR_KEYWORD,
        }

    def is_pos(self) -> bool:
        return self.parameter.kind in {
            Parameter.POSITIONAL_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.VAR_POSITIONAL,
        }

    def is_kw(self) -> bool:
        return self.parameter.kind in {
            Parameter.KEYWORD_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
            Parameter.VAR_KEYWORD,
        }

    def is_required(self) -> bool:
        return self.parameter.default == inspect.Parameter.empty and not self.is_var()


@attr.s(auto_attribs=True, frozen=True)
class _ParameterError(Exception):
    msg: str


def _get_parameters(type_: DeserializableType) -> List[_Parameter]:
    try:
        return [_Parameter(p) for p in inspect.signature(type_).parameters.values()]
    except ValueError as e:
        raise _ParameterError(str(e)) from e


def _get_type_hints(
    type_: DeserializableType, params: Sequence[inspect.Parameter]
) -> Dict[str, Any]:
    if hasattr(type_, "__globals__"):
        globals_ = getattr(type_, "__globals__")
    elif hasattr(type_, "__module__"):
        globals_ = vars(sys.modules[getattr(type_, "__module__")])
    else:  # pragma: no cover TODO
        globals_ = {}

    # For some reason `get_type_hints` can't handle the type returned by
    # `functools.partial` - but it _can_ handle its `__call__` method.
    if isinstance(type_, partial):
        type_ = type_.__call__

    try:
        type_hints = typing.get_type_hints(type_, globals_)
    except TypeError as e:  # pragma: no cover
        raise _ParameterError(str(e)) from None

    def get_type_hint(p: inspect.Parameter) -> Any:
        if p.name in type_hints:
            return type_hints[p.name]
        if p.annotation == inspect.Parameter.empty:
            return Any
        if not isinstance(p.annotation, str):
            return p.annotation
        # TODO __closure__?
        return eval(  # pragma: no cover, pylint: disable=eval-used
            p.annotation, globals_
        )

    return {p.name: get_type_hint(p) for p in params}



"""Automatically deserialize complex objects from simple Python types."""

import pkg_resources

from .errors import DeserializationError, DeserializerFactoryError
from .terramare import deserialize_into

# Attempt to expose the package version as __version__ (see PEP 396).
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:  # pragma: no cover
    pass


__all__ = ["DeserializationError", "DeserializerFactoryError", "deserialize_into"]



"""Deserializer implementing dynamic dispatch based on the value of a dictionary key."""

from functools import partial
from typing import Any, Generic, Hashable, Mapping, TypeVar, Union, cast

import attr

from . import error_utils
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Result
from .errors import DeserializerFactoryError
from .types import DeserializableType, Primitive, T, Tag

K = TypeVar("K", bound=Hashable)
S = TypeVar("S", bound=Hashable)


@attr.s(auto_attribs=True, frozen=True)
class KeyedDeserializer(Generic[K, S, T], Deserializer[T]):
    """Deserialize a dictionary based on the value of one of its keys."""

    _dict_ds: Deserializer[Mapping[K, Any]]

    _key: K
    _key_ds: Deserializer[S]

    _remainder_dss: Mapping[S, Deserializer[T]]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """Deserialize a dictionary based on the value of one of its keys."""
        elts_result = self._dict_ds.ds(cs, value)
        if elts_result.is_err():
            return cast(Result[T], elts_result)
        elts = elts_result.unwrap()

        if self._key not in elts:
            return Err(cs, f"missing key: '{self._key}'")

        key_result = self._key_ds.ds(cs.index(_key_str(self._key)), elts[self._key])
        if key_result.is_err():
            return cast(Result[T], key_result)
        key = key_result.unwrap()

        if key not in self._remainder_dss:
            return Err(
                cs.index(_key_str(self._key)),
                f"expected one of {{{', '.join(map(str, self._remainder_dss))}}}",
            )

        return self._remainder_dss[key].ds(
            cs.alternative(_key_str(key)),
            {k: v for k, v in elts.items() if k != self._key},
        )


@attr.s(auto_attribs=True, frozen=True)
class KeyedDeserializerFactory(Generic[K, S, T]):
    """Create a keyed deserializer."""

    def __call__(
        self,
        factory: DeserializerFactory,
        key: K,
        mapping: Mapping[S, DeserializableType],
    ) -> Deserializer[T]:
        """Create a KeyDeserializer from a specified key field and a value/type mapping."""
        return error_utils.unwrap(
            partial(DeserializerFactoryError, type_=Union[tuple(map(type, mapping))]),
            self.mk_ds_internal(factory, ContextStack.new(), key, mapping),
        )

    def mk_ds_internal(
        self,
        factory: DeserializerFactory,
        cs: ContextStack,
        key: K,
        mapping: Mapping[S, DeserializableType],
    ) -> Result[Deserializer[T]]:
        """
        Create a KeyDeserializer from a specified key field and mapping of key values to types.

        Recurse using the supplied factory.
        """
        return error_utils.collect_all(
            [
                factory.mk_ds(cs, Tag(Mapping[type(key), Any])),  # type: ignore[misc]
                factory.mk_ds(cs, Tag(Union[tuple(map(type, mapping))])),
            ]
            + [
                factory.mk_ds(cs.parameter_index(_key_str(k), t), Tag(t))
                for k, t in mapping.items()
            ]
        ).map(
            lambda dss: KeyedDeserializer(
                dss[0], key, dss[1], dict(zip(mapping, dss[2:]))
            )
        )


def _key_str(key: Any) -> Union[int, str]:
    if isinstance(key, int):
        return key
    return str(key)



"""Automatically deserialize complex objects from simple Python types."""

import dataclasses
from typing import (
    AbstractSet,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
)

from typing_extensions import Final, TypedDict

from terramare.stdlib_metadata import DEFAULT_METADATA, StdlibDeserializationError

from . import (
    classes,
    deserializer,
    enums,
    keyed_deserializer,
    mappings,
    metadata,
    newtypes,
    primitives,
    sequences,
    type_utils,
    unions,
)
from .types import DeserializableType, Primitive, T, TypedDictMetas

__all__ = [
    "DEFAULT_FALSE_STRINGS",
    "DEFAULT_TRUE_STRINGS",
    "Primitive",
    "create_deserializer_factory",
    "deserialize_into",
]

DEFAULT_TRUE_STRINGS: Final = frozenset({"yes", "on", "true", "1"})
DEFAULT_FALSE_STRINGS: Final = frozenset({"no", "off", "false", "0"})


def deserialize_into(
    type_: DeserializableType,
    value: Primitive,
    *,
    coerce_strings: bool = False,
    true_strings: AbstractSet[str] = DEFAULT_TRUE_STRINGS,
    false_strings: AbstractSet[str] = DEFAULT_FALSE_STRINGS,
    handle_exception_types: AbstractSet[Type[Exception]] = frozenset(),
    _experimental_metadata: Optional[
        Mapping[DeserializableType, Optional[metadata.Metadata]]
    ] = None,
) -> T:
    """
    Deserialize a primitive as a value of the specified type.

    :param `type_`: Deserialize into this type.
    :param `value`: Primitive value to attempt to deserialize.
    :param `coerce_strings`: If set, attempt to convert :python:`str` values to
        :python:`bool`, :python:`int`, or :python:`float` where the latter are required.
        For example:

        >>> deserialize_into(int, "1")
        Traceback (most recent call last):
            ...
        terramare.errors.DeserializationError: ...
        >>> deserialize_into(int, "1", coerce_strings=True)
        1

        Note that setting this option will cause :python:`terramare` to reject
        non-string primitives where a :python:`bool`, :python:`int`, or :python:`float`
        is required.
        For example:

        >>> deserialize_into(int, 1)
        1
        >>> deserialize_into(int, 1, coerce_strings=True)
        Traceback (most recent call last):
            ...
        terramare.errors.DeserializationError: ...

        This option defaults to :python:`False`.

    :param `true_strings`: Set of strings to convert to :python:`True` when convering a
        :python:`str` value to a :python:`bool`. Case is ignored.

        This value defaults to :python:`{"yes", "on", "true", "1"}`.

    :param `false_strings`: Set of strings to convert to :python:`False` when convering
        a :python:`str` value to a :python:`bool`. Case is ignored.

        This value defaults to :python:`{"no", "off", "false", "0"}`.

    :param `handle_exception_types`: Set of additional exception types that
        :python:`terramare` should catch and handle rather than propogating.

        Generally this will still result in an exception being raised. However, it will
        be a :python:`terramare` exception containing additional context.

        This option is useful when the deserialization target provides some form of
        additional validation. For example:

        >>> from typing import Union
        >>> import attr
        >>>
        >>> @metadata.from_(metadata.VALUE)
        ... @attr.s
        ... class Paint:
        ...    color: str = attr.ib(validator=attr.validators.in_(["red", "blue"]))

        When no exception types are provided, context is lost when deserializing into a
        single type:

        >>> deserialize_into(Paint, "green")
        Traceback (most recent call last):
        ...
        ValueError: 'color' must be in ['red', 'blue'] (got 'green')

        When the exception type is provided, that context is preserved:

        >>> deserialize_into(Paint, "green", handle_exception_types={ValueError})
        Traceback (most recent call last):
        ...
        terramare.errors.DeserializationError: ...
        .: no alternative matched
        └─ • value: 'color' must be in ['red', 'blue'] (got 'green')

        The exception is when the deserialization target is a union type, where
        :python:`terramare` will continue to try further union variants - as it would
        when encountering a deserialization failure.

        When no exception types are provided, :python:`terramare` cannot try further
        union variants if one fails due to a validation error:

        >>> deserialize_into(Union[Paint, str], "green")
        Traceback (most recent call last):
        ...
        ValueError: 'color' must be in ['red', 'blue'] (got 'green')

        When the exception type is provided, :python:`terramare` continues and
        successfully deserializes into a later variant.

        >>> deserialize_into(Union[Paint, str], "green", handle_exception_types={ValueError})
        'green'

    :param _experimental_metadata: See :ref:`experimental-features-metadata`.

    :raises terramare.DeserializerFactoryError: if a deserializer for :python:`type_`
        cannot be created.
    :raises terramare.DeserializationError: if the deserializer fails to deserialize a
        value of :python:`type_` from :python:`value`.
    """  # noqa: E501

    return create_deserializer_factory(
        coerce_strings=coerce_strings,
        true_strings=true_strings,
        false_strings=false_strings,
        handle_exception_types=handle_exception_types,
        _experimental_metadata=_experimental_metadata,
    )(type_)(value)


def create_deserializer_factory(
    *,
    coerce_strings: bool = False,
    true_strings: AbstractSet[str] = DEFAULT_TRUE_STRINGS,
    false_strings: AbstractSet[str] = DEFAULT_FALSE_STRINGS,
    handle_exception_types: AbstractSet[Type[Exception]] = frozenset(),
    _experimental_metadata: Optional[
        Mapping[DeserializableType, Optional[metadata.Metadata]]
    ] = None,
) -> deserializer.DeserializerFactory:
    """Create a DeserializerFactory using sensible defaults which may be overridden."""
    metadata_collection = metadata.MetadataCollection(
        {
            k: v
            for k, v in {**DEFAULT_METADATA, **(_experimental_metadata or {})}.items()
            if v is not None
        }
    )

    def _class_deserializer_enable_if(t: DeserializableType) -> metadata.FromType:
        # Explicitly disable class deserializer creation for typing types.
        # They are technically callable, so the class deserializer factory will
        # create a deserializer, but this will always raise if called.
        if type_utils.get_base_of_generic_type(t) in {
            AbstractSet,
            Callable,
            Dict,
            FrozenSet,
            Iterable,
            Iterator,
            List,
            Mapping,
            MutableMapping,
            MutableSequence,
            Sequence,
            Set,
            Tuple,
            TypedDict,
            *TypedDictMetas,
        }:
            return frozenset()
        try:
            t_metadata = metadata_collection[t]
        except KeyError:
            pass
        else:
            # Metadata of other types will be consumed by the
            # MetadataDeserializerFactory before reaching this function, making
            # this branch hard to miss.
            if isinstance(  # pragma: no branch
                # pylint: disable=protected-access
                t_metadata,
                metadata.From,
            ):
                return t_metadata.from_
        if _is_attrs_class(t) or _is_dataclass(t):
            return metadata.DICT
        if _is_namedtuple(t):
            return metadata.DICT | metadata.LIST
        return frozenset()

    return deserializer.OneOfDeserializerFactory(
        {
            "metadata": metadata.MetadataDeserializerFactory(
                metadata_collection, keyed_deserializer.KeyedDeserializerFactory()
            ),
            "newtype": newtypes.NewTypeDeserializerFactory(),
            "primitive": primitives.PrimitiveDeserializerFactory(
                {
                    **{k.lower(): True for k in true_strings},
                    **{k.lower(): False for k in false_strings},
                }
                if coerce_strings
                else None
            ),
            "literal": enums.LiteralDeserializerFactory(),
            "enum": enums.EnumDeserializerFactory(),
            "union": unions.UnionDeserializerFactory(),
            "tuple": sequences.TupleDeserializerFactory(),
            "sequence": sequences.HomogeneousSequenceDeserializerFactory(),
            "typeddict": mappings.TypedDictDeserializerFactory(),
            "dict": mappings.DictDeserializerFactory(),
            "class": classes.ClassDeserializerFactory(
                enable_if=_class_deserializer_enable_if,
                handle_exception_ts=handle_exception_types
                | {StdlibDeserializationError},
            ),
        }
    )


def _is_attrs_class(t: DeserializableType) -> bool:
    return hasattr(t, "__attrs_attrs__")


def _is_dataclass(t: DeserializableType) -> bool:
    return isinstance(t, type) and dataclasses.is_dataclass(t)


def _is_namedtuple(t: DeserializableType) -> bool:
    return (
        isinstance(t, type)
        and issubclass(t, tuple)
        and hasattr(t, "_asdict")
        and hasattr(t, "__annotations__")
    )



"""Deserializers for typed dictionaries."""

from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    TypeVar,
    cast,
)

import attr
from typing_extensions import TypedDict

from . import error_utils, iterator_utils, type_utils
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .types import DeserializableType, Primitive, T, Tag

K = TypeVar("K", bound=Hashable)


@attr.s(auto_attribs=True, frozen=True)
class MappingDeserializer(Generic[K, T], Deserializer[T]):
    """Deserialize a dictionary using a mapping of keys to deserializers."""

    _untyped_ds: Deserializer[Dict[Any, Any]]

    _key_ds: Deserializer[K]
    _required_dss: Mapping[K, Deserializer[Any]]
    _optional_dss: Mapping[K, Deserializer[Any]]
    _var_ds: Optional[Deserializer[Any]]

    _construct_fn: Callable[[ContextStack, Mapping[K, Any]], Result[T]]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """
        Deserialize a primitive using a mapping of keys to deserializers.

        Each dictionary element is matched with a deserializer based on its key.
        An error will be raised if the dictionary is missing keys corresponding
        to required element deserializers.
        An error will also be raised if the dictionary contains keys that do not
        correspond to either required or optional element deserializers, and
        no variadic element deserializer is supplied.
        """
        untyped_elts = self._untyped_ds.ds(cs, value)
        if untyped_elts.is_err():
            return cast(Result[T], untyped_elts)

        keys = error_utils.collect_all(
            partial(self._key_ds.ds, cs.index(i), k)
            for i, k in enumerate(untyped_elts.unwrap())
        )
        if keys.is_err():
            return cast(Result[T], keys)

        try:
            zipped = iterator_utils.zip_strict_dict(
                self._required_dss,
                self._optional_dss,
                self._var_ds,
                dict(zip(keys.unwrap(), untyped_elts.unwrap().values())),
            )
        except ValueError as e:
            return Err(cs, str(e))
        # pylint: disable=protected-access
        result = error_utils.collect_all(
            partial(ds.ds, cs.index(str(k)), v) for k, (ds, v) in zipped.items()
        )
        if result.is_ok():
            return self._construct_fn(cs, dict(zip(zipped.keys(), result.unwrap())))
        return cast(Result[T], result)


@attr.s(auto_attribs=True, frozen=True)
class DictDeserializerFactory(DeserializerFactory):
    """Create deserializer for a dictionary of values of the same type."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified dictionary type."""
        if type_utils.get_base_of_generic_type(type_.t) not in {
            Dict,
            Mapping,
            MutableMapping,
        }:
            return Err(cs, "type mismatch")

        key_t, value_t = type_utils.get_type_parameters(type_.t)
        if not _is_hashable(key_t):
            return Ok(Err(cs.type_index(key_t), "unhashable type"))

        return Ok(
            error_utils.collect_all(
                [
                    factory.mk_ds(cs, Tag(dict)),
                    factory.mk_ds(cs.type_index(key_t), Tag(key_t)),
                    factory.mk_ds(cs.type_index(value_t), Tag(value_t)),
                ]
            ).map(
                lambda dss: MappingDeserializer(
                    dss[0], dss[1], {}, {}, dss[2], lambda _, v: Ok(cast(T, dict(v)))
                )
            )
        )


@attr.s(auto_attribs=True, frozen=True)
class TypedDictDeserializerFactory(DeserializerFactory):
    """Create deserializer for a TypedDict."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified TypedDict type."""
        if type_utils.get_base_of_generic_type(type_.t) != TypedDict:
            return Err(cs, "type mismatch")

        annotations = type_.t.__annotations__
        return Ok(
            error_utils.collect_all(
                [factory.mk_ds(cs, Tag(dict)), factory.mk_ds(cs, Tag(str))]
                + [
                    factory.mk_ds(cs.parameter_index(k, t), Tag(t))
                    for k, t in annotations.items()
                ],
            ).map(
                lambda dss: MappingDeserializer(
                    dss[0],
                    dss[1],
                    dict(zip(annotations, dss[2:])),
                    {},
                    None,
                    lambda _, v: Ok(cast(T, v)),
                )
            )
        )


def _is_hashable(type_: DeserializableType) -> bool:
    return (
        "__hash__" in dir(type_)
        and "__eq__" in dir(type_)
        and type_utils.get_base_of_generic_type(type_) not in {Dict, List, Set}
    )
